import json
import os
from datetime import datetime

from meross_iot.cloud.exceptions.CredentialsException import InvalidStoredCredentialsException


class MerossCloudCreds(object):
    def __init__(self, token, key, user_id, user_email, issued_on):
        self.token = token
        self.key = key
        self.user_id = user_id
        self.user_email = user_email
        self.issued_on = issued_on

    def serialize_key_token(self, filepath):
        d = {
            'token': self.token,
            'key': self.key,
            'user_id': self.user_id,
            'user_email': self.user_email,
            'issued_on': self.issued_on.isoformat()
        }

        with open(file=filepath, mode="wb+") as f:
            json.dump(d, f)

    @staticmethod
    def load_from_json_file(filepath):
        if not os.path.isfile(filepath):
            raise InvalidStoredCredentialsException("Invalid file path: %s" % filepath)

        with open(file=filepath, mode="rb") as f:
            data = json.load(f)
            token = data.get('token')
            key = data.get('key')
            user_id = data.get('user_id')
            user_email = data.get('user_email')
            issued_on = data.get('issued_on')

            if token is None:
                raise InvalidStoredCredentialsException("The stored credentials are not valid. Missing token attribute.")
            if key is None:
                raise InvalidStoredCredentialsException("The stored credentials are not valid. Missing key attribute.")
            if user_id is None:
                raise InvalidStoredCredentialsException("The stored credentials are not valid. Missing user_id attribute.")
            if user_email is None:
                raise InvalidStoredCredentialsException("The stored credentials are not valid. Missing user_email attribute.")
            if issued_on is None:
                raise InvalidStoredCredentialsException("The stored credentials are not valid. Missing issued_on attribute.")
            else:
                issued_on = datetime.fromisoformat(issued_on)

            return MerossCloudCreds(token=token, key=key, user_id=user_id, user_email=user_email, issued_on=issued_on)



import base64
import hashlib
import json
import random
import string
import time
from enum import Enum

import requests

from meross_iot.cloud.exceptions.TokenExpiredException import TokenExpiredException
from meross_iot.cloud.exceptions.TooManyTokensException import TooManyTokensException
from meross_iot.credentials import MerossCloudCreds
from meross_iot.logger import API_LOGGER as l
from datetime import datetime


# Appears to be used as a part of the signature algorithm as constant "salt" (kinda useless)
_SECRET = "23x17ahWarFH6w29"
_MEROSS_URL = "https://iot.meross.com"
_LOGIN_URL = "%s%s" % (_MEROSS_URL, "/v1/Auth/Login")
_LOG_URL = "%s%s" % (_MEROSS_URL, "/v1/log/user")
_DEV_LIST = "%s%s" % (_MEROSS_URL, "/v1/Device/devList")
_HUB_DUBDEV_LIST = "%s%s" % (_MEROSS_URL, "/v1/Hub/getSubDevices")
_LOGOUT_URL = "%s%s" % (_MEROSS_URL, "/v1/Profile/logout")


class ErrorCodes(Enum):
    CODE_NO_ERROR = 0
    CODE_TOKEN_EXPIRED = 1019
    CODE_TOO_MANY_TOKENS = 1301


# TODO: implement Token check API call (if any)
#       and automatic token renewal
class MerossHttpClient:
    def __init__(self, cloud_credentials):
        self._cloud_creds = cloud_credentials

    @classmethod
    def from_user_password(cls, email, password):
        creds = cls.login(email, password)
        return MerossHttpClient(cloud_credentials=creds)

    @classmethod
    def login(cls, email, password):
        data = {"email": email, "password": password}
        response_data = cls._authenticated_post(_LOGIN_URL, params_data=data)
        creds = MerossCloudCreds(
            token=response_data["token"],
            key=response_data["key"],
            user_id=response_data["userid"],
            user_email=response_data["email"],
            issued_on=datetime.utcnow()
        )
        return creds

    @classmethod
    def _authenticated_post(cls,
                            url,  # type: str
                            params_data,  # type: dict,
                            cloud_creds=None
                            ):

        nonce = cls._generate_nonce(16)
        timestamp_millis = int(round(time.time() * 1000))
        login_params = cls._encode_params(params_data)

        # Generate the md5-hash (called signature)
        m = hashlib.md5()
        datatosign = '%s%s%s%s' % (_SECRET, timestamp_millis, nonce, login_params)
        m.update(datatosign.encode("utf8"))
        md5hash = m.hexdigest()

        headers = {
            "Authorization": "Basic" if cloud_creds is None else "Basic %s" % cloud_creds.token,
            "vender": "Meross",
            "AppVersion": "1.3.0",
            "AppLanguage": "EN",
            "User-Agent": "okhttp/3.6.0"
        }

        payload = {
            'params': login_params,
            'sign': md5hash,
            'timestamp': timestamp_millis,
            'nonce': nonce
        }

        # Perform the request.
        r = requests.post(url, data=payload, headers=headers)

        # Check if that is ok.
        if r.status_code != 200:
            raise AuthenticatedPostException("Failed request to API. Response code: %d" % str(r.status_code))

        # Save returned value
        jsondata = r.json()
        code = jsondata.get('apiStatus')

        if jsondata["info"].lower() != "success":
            l.warn("Failed request to API. Response was: %s" % str(jsondata))
            raise AuthenticatedPostException("Failed request to API. Response was: %s" % str(jsondata))

        try:
            result_code = ErrorCodes(code)
            if result_code == ErrorCodes.CODE_NO_ERROR:
                return jsondata["data"]
            elif result_code == ErrorCodes.CODE_TOKEN_EXPIRED:
                raise TokenExpiredException("The provided token has expired")
            elif result_code == ErrorCodes.CODE_TOO_MANY_TOKENS:
                raise TooManyTokensException("You have issued too many tokens without logging out.")
            else:
                raise AuthenticatedPostException("Failed request to API. Response was: %s" % str(jsondata))
        except:
            raise AuthenticatedPostException("Cannot parse response code. Response was: %s" % str(jsondata))

    @classmethod
    def _encode_params(cls,
                       parameters  # type: dict
                       ):
        jsonstring = json.dumps(parameters)
        return str(base64.b64encode(jsonstring.encode("utf8")), "utf8")

    @classmethod
    def _generate_nonce(cls, length):
        return ''.join(random.SystemRandom().choice(string.ascii_uppercase + string.digits) for _ in range(length))

    def _log(self):
        data = {'extra': {}, 'model': 'Android,Android SDK built for x86_64', 'system': 'Android',
                'uuid': '493dd9174941ed58waitForOpenWifi', 'vendor': 'Meross', 'version': '6.0'}
        self._authenticated_post(_LOG_URL, params_data=data)

    def list_devices(self):
        return self._authenticated_post(_DEV_LIST, {}, cloud_creds=self._cloud_creds)

    def get_cloud_credentials(self):
        return self._cloud_creds

    def list_hub_subdevices(self, hub_id):
        return self._authenticated_post(_HUB_DUBDEV_LIST, {"uuid": hub_id}, cloud_creds=self._cloud_creds)

    def logout(self):
        return self._authenticated_post(_LOGOUT_URL, {}, cloud_creds=self._cloud_creds)


class AuthenticatedPostException(Exception):
    pass


class UnauthorizedException(Exception):
    pass



import os
import threading
import time
import logging
import sys, traceback
from uuid import uuid4

l = logging.getLogger(__name__)


class LockFactory(object):
    def __init__(self, enable_debug=False):
        self._debug = enable_debug
        self._lock_timeout_threshold = 5

    def configure_factory(self, enable_debug, lock_timeout_threshold):
        self._debug = enable_debug
        self._lock_timeout_threshold = lock_timeout_threshold

    def build_rlock(self):
        if self._debug:
            return ManagedRLock(lock_timeout_threshold=self._lock_timeout_threshold)
        else:
            return threading.RLock()


class LockMap(object):
    def __init__(self):
        self._lock_dict = {}
        self._lock = threading.RLock()

    def register_lock(self, lock_uuid):
        with self._lock:
            # { uuid: [lock_uuid, thread, count]}
            lock_data = self._lock_dict.get(lock_uuid, [lock_uuid, threading.current_thread(), 0])
            lock_data[2] += 1
            self._lock_dict[lock_uuid] = lock_data

    def unregister_lock(self, lock_uuid):
        with self._lock:
            # { uuid: [uuid, count]}
            lock_data = self._lock_dict.get(lock_uuid, None)
            if lock_data is None:
                raise Exception("ILLEGAL STATE")
            lock_data[2] -= 1
            if lock_data[2] == 0:
                del self._lock_dict[lock_uuid]
            else:
                self._lock_dict[lock_uuid] = lock_data

    def get_locking_thread_uuid(self, lock_uuid):
        with self._lock:
            return self._lock_dict.get(lock_uuid, [None, None, None])[1]


class ManagedRLock(object):
    def __init__(self, lock_timeout_threshold):
        # This is the guardina lock, used to synchronize access to this lock
        self._guad_lock = threading.RLock()

        self._lock_timeout_threshold = lock_timeout_threshold
        self._lock = threading.RLock()
        self._uuid = str(uuid4())

        try:
            self._is_owned = self._lock._is_owned
        except AttributeError:
            pass

    def acquire(self, blocking=True, timeout=-1):
        with self._guad_lock:
            result = self._lock.acquire(blocking=False)
            if not result and lockmap.get_locking_thread_uuid(self._uuid) is not None:
                l.warning("Possible DEADLOCK. Thread %s is trying to acquire lock %s, "
                          "which is already owned by thread %s." % (threading.current_thread().name, self._uuid,
                                                                    lockmap.get_locking_thread_uuid(self._uuid).name))
            elif result:
                self._lock.release()

        result = self._lock.acquire(blocking=blocking, timeout=timeout)
        if result:
            lockmap.register_lock(self._uuid)

        return result

    def release(self):
        with self._guad_lock:
            result = self._lock.release()
            lockmap.unregister_lock(self._uuid)
            return result

    def __enter__(self):
        return self.acquire()

    def __exit__(self, *args):
        return self.release()

    def __repr__(self):
        return "<ManagedRLock(%s)>" % self._uuid


# LockFactory Singleton
lock_factory = LockFactory()
lockmap = LockMap()


if __name__ == '__main__':
    root = logging.getLogger()
    root.setLevel(logging.DEBUG)
    handler = logging.StreamHandler(sys.stdout)
    handler.setLevel(logging.DEBUG)
    formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
    handler.setFormatter(formatter)
    root.addHandler(handler)

    # Test with a deadlock
    l1 = ManagedRLock(lock_timeout_threshold=5)
    l2 = ManagedRLock(lock_timeout_threshold=5)

    def deadlocker(l1, l2):
        with l1:
            with l2:
                print("Lock acquired.")
                time.sleep(120)

    t1 = threading.Thread(target=deadlocker, args=(l1,l2))
    t2 = threading.Thread(target=deadlocker, args=(l2, l1))
    t1.start()
    t2.start()
    t2.join()
    t1.join()
